/*
 * D64VRM (/) 2010 iAN CooG/HVSC
 * Find HIV/BHP/BULA infected D64s and clean them by restoring the t/s links
 * based on d642prg (Original (C) Covert Bitops, mod by iAN CooG)
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define VERSION "0.3"
#define DIRSECTS  18
#define ArrayLen(x) (sizeof(x)/sizeof(*x))

int d64bytes[4]={ 174848  /*35 no err*/
                , 175531  /*35 w/ err*/
                , 196608  /*40 no err*/
                , 197376  /*40 w/ err*/
                };

int snumtable[] =
{
    21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,
    19,19,19,19,19,19,19,
    18,18,18,18,18,18,
    17,17,17,17,17
   ,17,17,17,17,17 /*tracks 36-40*/
};

int firstsecttbl[40];

/*****************************************************************************/
int getoffset(int track, int sector)
{
    int offset;
    track--;
    offset = (firstsecttbl[track]+sector)*256;
    return offset;
}

/*****************************************************************************/
int main(int argc, char **argv)
{
    int c,s,VirFound=0,skipblx=0,retval=0;
    int ofs,imgsize,totaltracks=0;
    int tr=0,sc=0,blk=0,nt=0;
    char parsepar=1;
    char vfclean= 1;
    char vfverb = 0;
    char vIonly = 0;
    unsigned char prgfilename[20]={0};
    char fd64[260]={0};
    char flnm[24]={0};
    unsigned char *p;
    unsigned char csplat;
    unsigned char cprot;
    char fln2[20]={0};
    char types[][6]={"DEL","SEQ","PRG","USR","REL","???"};
    FILE *d64handle=NULL;
    unsigned char *d64buf;
    unsigned char *ptr,*fptr;
    unsigned char dirsectors[DIRSECTS+1]={0};

    for (c = 1; c < argc; c++)
    {
        if ( (parsepar==1) && (argv[c][0] == '-') )
        {
            switch(argv[c][1])
            {
            case 's':
            case 'S':
                vfclean = 0;
                break;
            case 'v':
            case 'V':
                vfverb  = 1;
                break;
            case 'i':
            case 'I':
                vIonly  = 1;
                break;
            case '-':
                parsepar=0;
                break;
            }
        }
        else
            strcpy(fd64,argv[c]);
    }

    if((argc == 1)||!*fd64||(vfverb&&vIonly))
    {
        printf( "D64 Virus ReMover v"VERSION" (C) 2010-2011 iAN CooG/HVSC\n\n"
                "Usage: d64vrm [-s] [-v/-i] [--] <d64 image>\n"
                "-s to Scan only\n"
                "-v for Verbose output / -i output only if d64 is infected\n"
                "-- stop parsing parameters (use if filename starts with -)\n"
              );
        return 255;
    }


    d64handle = fopen(fd64, "rb");
    if (!d64handle)
    {
        if(!vIonly)
            printf("Couldn't open d64 image %s.\n",fd64);
        return 10;
    }
    fseek(d64handle,0,SEEK_END);
    imgsize=ftell(d64handle);
    fseek(d64handle,0,SEEK_SET);
    d64buf=NULL;
    for(c=0;c<ArrayLen(d64bytes);c++)
    {
        if(d64bytes[c]==imgsize)
        {
            if(c<2)
                totaltracks=35;
            else
                totaltracks=40;
            break;
        }
    }
    if(c==ArrayLen(d64bytes))
    {
        if(!vIonly)
            printf("Not a valid d64 image %s (%d bytes).\n",fd64,imgsize);
        fclose(d64handle);
        return 11;
    }
    d64buf = malloc(imgsize);
    if (!d64buf)
    {
        if(!vIonly)
            printf("No memory for d64 image %s.\n",fd64);
        fclose(d64handle);
        return 12;
    }


    fread(d64buf,imgsize,1,d64handle);
    fclose(d64handle);

    s = 0;
    for (c = 0; c < totaltracks; c++)
    {
        firstsecttbl[c] = s;
        s += snumtable[c];
    }

    if(vfverb)
    {
        p=flnm;
        ptr = &d64buf[getoffset(18,0)];
        strncpy(flnm,&ptr[0x90],23);
        *(p+16)='\"';
        *(p+17)='\"';
        while(*p)
        {
            if (*p == 0xa0)
                *p = 0x20;
            p++;
        }

        printf("\n0 \"%s\"\n"
               "----- ------------------  ----- -----",flnm);
    }
    ptr = &d64buf[getoffset(18,1)];
    dirsectors[0]=(unsigned char)0x01;
    dirsectors[DIRSECTS] = dirsectors[0];
    ofs = 2;

    for (;;)
    {
        memcpy(prgfilename,&ptr[ofs+3],16);
        prgfilename[16]=0;
        for (c =0;c<=15 ;c++)  /* fix 0s in filename -> a0 */
        {
            if (prgfilename[c] == 0)
                prgfilename[c] = 0xa0;
        }
        for (c =15; c>=0 ;c--)  /* trailing a0 -> 0 */
        {
            if (prgfilename[c] == 0xa0)
                prgfilename[c] = 0;
            else
                break;
        }

        /* convert filename */

        for (c = 0; c<16;c++)
        {
            if( prgfilename[c] == 0xc0)
                prgfilename[c] = '-';
            else if( prgfilename[c] >  0x80)
                prgfilename[c] -= 0x80;

            if( (prgfilename[c]) && (prgfilename[c] <= 0x1f))
                prgfilename[c] += 0x60;

            if( /*ptr[ofs+3+c] >=  0x80 ||*/
                prgfilename[c] == '?'   ||
                prgfilename[c] == '*'   ||
                prgfilename[c] == '\\'  ||
                prgfilename[c] == '/'   ||
                prgfilename[c] == '%'   ||
                prgfilename[c] == '<'   ||
                prgfilename[c] == '>'   ||
                prgfilename[c] == ':'   ||
                prgfilename[c] == '^'   ||
                prgfilename[c] == '\x7f'||
                prgfilename[c] == '|')
                prgfilename[c] = '_';
        }

        csplat=' ';
        cprot=' ';
        tr=0;
        sc=0;
        blk=0;
        nt=0;

        strcpy(fln2, "\"");
        strncat(fln2, prgfilename,16);
        fln2[17]=0; // just in case c64 filename has garbage after 16th char, like GEOS files
        strcat(fln2, "\"");
        sprintf(flnm,"%-19s", fln2);
        tr= (int)(unsigned short int)(ptr[ofs+1]);
        sc= (int)(unsigned short int)(ptr[ofs+2]);
        blk=(int)(unsigned short int)(ptr[ofs+28]|(ptr[ofs+29]<<8));

        nt= ptr[ofs] & 7;
        if(nt>4) nt=5;
        //if(ptr[ofs]) // to display really deleted files as "*DEL", $80 as " DEL"
        {
            if( (ptr[ofs] & 0x80) == 0)
            {
                csplat='*';
            }
            if( (ptr[ofs] & 0x40) )
            {
                cprot='<';
            }
        }
        if(vfverb&&(ptr[ofs+3]||tr||sc||blk)) /* skip empty entries*/
        {
            printf("\n%5d %s %c%s%c %02d;%02d",
                    blk,flnm,csplat,types[nt],cprot,tr,sc);
        }

        if ((ptr[ofs] & 7)<6)
        {
            if((ptr[ofs+1]<1)||(ptr[ofs+1]>totaltracks)||(ptr[ofs+2]>=snumtable[(int)(ptr[ofs+1]-1)]))
            {
                //printf("Invalid track %d sector %d\n", ptr[ofs+1],ptr[ofs+2]);
                break;
            }
            fptr = &d64buf[getoffset(ptr[ofs+1],ptr[ofs+2])];

            if ((ptr[ofs] & 7)&&fptr[0]) /* skip deleted files and 1 block files */
            {
                if(fptr[8]==0x9e)
                {
                    if((*(unsigned int *)(fptr+0x09)==0x39353032)&&
                       (*(unsigned int *)(fptr+0x10)==0xA94910DD)&&
                       (*(unsigned int *)(fptr+0xa3)==0x4D2D57A0)&&
                       (*(unsigned int *)(fptr+0xe3)==0x054F4CC3) )
                    {
                        VirFound++;
                        skipblx=2;
                        if(vfverb)
                        {
                            printf(" HIV ");
                        }
                    }
                    else
                    if((*(unsigned int *)(fptr+0x34)==0x2BB122A0)&&
                       (*(unsigned int *)(fptr+0x38)==0x2AC0C848)&&
                       (*(unsigned int *)(fptr+0xd2)==0xBD09A20B)&&
                       (*(unsigned int *)(fptr+0xe2)==0xFFFBAD48) )
                    {
                        VirFound++;
                        skipblx=8;
                        if(vfverb)
                        {
                            printf(" BHP ");
                        }
                    }
                    else
                    if((*(unsigned int *)(fptr+0x10)==0x206D627B)&&
                       (*(unsigned int *)(fptr+0x20)==0x404CF7D0)&&
                       (*(unsigned int *)(fptr+0x60)==0x57A9FFA8)&&
                       (*(unsigned int *)(fptr+0x70)==0x2020A9FF) )
                    {
                        VirFound++;
                        skipblx=3;
                        if(vfverb)
                        {
                            printf(" BULA ");
                        }
                    }
                    if( skipblx )
                    {
                        if(vfclean)
                        {
                            for(c=0;c<skipblx;c++)
                            {
                                tr=fptr[0];
                                sc=fptr[1];
                                memset(fptr,0,0x100);
                                fptr = &d64buf[getoffset(tr,sc)];
                            }

                            ptr[ofs+1] = tr;
                            ptr[ofs+2] = sc;
                            if(blk>skipblx)
                                blk-=skipblx;
                            else
                                blk=0;
                            ptr[ofs+28]=(blk&0xff);
                            ptr[ofs+29]=(blk>>8);
                            if(vfverb)
                            {
                                printf("CLEANED");
                            }
                        }
                        else
                        {
                            if(vfverb)
                            {
                                printf("FOUND");
                            }
                        }
                        skipblx=0;

                    }
                }
            }
        }
        /* next dir entry */
        ofs += 32;
        if (ofs >= 256)
        {   /* next dir sector */
            if (ptr[0]==0x12/*&&ptr[1]>0&&ptr[1]<0xff*/) // next dir sectors only on trk 18 and NEVER on sector 0
            {
                for(c=0;c<DIRSECTS;c++)
                {
                    if(dirsectors[c]==0)     // one slot free, add the next and parse it
                    {
                        dirsectors[c]=ptr[1];
                        dirsectors[DIRSECTS] = ptr[1];
                        break;
                    }
                    if(dirsectors[c]==ptr[1])// this sector was already parsed, recursive directory!
                    {
                        c=-1;
                        if(!vIonly)
                            printf("      RECURSIVEDIR:%02d>%02d\n",dirsectors[DIRSECTS],ptr[1]);
                        break;
                    }
                }
                if(c==-1)
                    break;
            }
            else
            {
                if( ptr[0] )
                {
                    if(!vIonly)
                        printf("      INVALID LINK:%02d;%02d\n",ptr[0],ptr[1]);
                }
                break;
            }

            if (ptr[0])
            {
                ptr = &d64buf[getoffset(ptr[0],ptr[1])];
                ofs = 2;
            }
            else
            {
                break;
            }
        }
    }
    if(vfverb)
    {
        printf("\n");
    }

    if(VirFound)
    {
        printf("%s",fd64);
        if(vfclean)
        {
            d64handle = fopen(fd64, "wb");
            if (!d64handle)
            {
                printf(" Couldn't rewrite d64 image\n");
                return 4;
            }
            fwrite(d64buf,imgsize,1,d64handle);
            fclose(d64handle);
            printf(" cleaned\n");
            retval=2;
        }
        else
        {
            printf(" INFECTED\n");
            retval=1;
        }
    }
    else
    {
        if(!vIonly)
        {
            printf("%s",fd64);
            printf(" clean\n");
        }
        retval=0;
    }
    return retval;
}

